<?php
/**
*
* @package extender: core and template event listeners constructor
* @copyright (c) 2014 c61 c61@yandex.ru http://c61.su
* @license http://opensource.org/licenses/gpl-2.0.php GNU General Public License v2
*
*/
namespace c61\extender\event;

use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\EventDispatcher\ContainerAwareEventDispatcher;

/**
* Event listener
*/
class listener implements EventSubscriberInterface
{	
	/** @var \phpbb\auth\auth */
	protected $auth;
	
	/** @var \phpbb\config\config */
	protected $config;
	
	/** @var \phpbb\config\db_text */
	protected $config_text;
	
	/** @var \phpbb\request\request_interface */
	protected $request;

	/** @var \phpbb\template\template */
	protected $template;
	
	/** @var \phpbb\user */
	protected $user;
	
	/** @var \phpbb\db\driver\driver_interface */
	protected $db;
	
	/** @var string */
	protected $phpbb_root_path;
	protected $php_ext;

	// Extender vars

	// rootref
	private $rootref;
	// user lang
	private $userlang;
	// template data
	private $tpldata;
	// event object
	private $event;

	// php-code for eval
	private $eval;
	// saved vars
	private $saved_vars = array();

	// allow php
	private $allow_php = true;

	// Core events collection
	static private $core_events = array();

	/**
	* Constructor
	* 
	* @param \phpbb\auth\auth $auth
	* @param \phpbb\config\config $config
	* @param \phpbb\config\db_text $config_text
	* @param \phpbb\template\template $template
	* @param \phpbb\user $user
	* @param \phpbb\db\driver\driver_interface $db
	* @param string $phpbb_root_path Root path
	* @param string $php_ext
	*/
	
	public function __construct(\phpbb\auth\auth $auth, \phpbb\config\config $config, \phpbb\config\db_text $config_text, \phpbb\request\request_interface $request, \phpbb\template\template $template, \phpbb\user $user, \phpbb\db\driver\driver_interface $db, $phpbb_root_path, $php_ext)
	{
		$this->auth = $auth;
		$this->config = $config;
		$this->config_text = $config_text;
		$this->request = $request;
		$this->template = $template;
		$this->user = $user;
		$this->db = $db;
		$this->phpbb_root_path = $phpbb_root_path;
		$this->php_ext = $php_ext;
	}
	
	/**
	* Assign functions defined in this class to event listeners in the core
	*
	* @return array
	* @static
	* @access public
	*/
	static public function getSubscribedEvents()
	{
		global $phpbb_container;

		$ret = array();
		$eindex = 0;

		$config = $phpbb_container->get('config');
		if ( isset($config['extender_core_on']) && $config['extender_core_on'] )
		{
			$config_text = $phpbb_container->get('config_text');
			$extender_core_data_prepared = $config_text->get_array(array('extender_core_prepared'));
			$extender_core_prepared = isset($extender_core_data_prepared['extender_core_prepared']) ? $extender_core_data_prepared['extender_core_prepared'] : null;

			if ( !empty($extender_core_prepared) )
			{
				self::$core_events = unserialize($extender_core_prepared);
				foreach (self::$core_events as $event)
				{
					if ( !isset($event['name']) ) break;
					$ret[$event['name']][] = array('event_extender_core_'.$eindex,$event['priority']);
					$eindex++;
				}
			}
			else
			{
				$extender_core_data = $config_text->get_array(array('extender_core'));
				$extender_core = htmlspecialchars_decode($extender_core_data['extender_core']);
				if ( !empty($extender_core)
				  && preg_match_all('/<!--\s*EVENT\s*([-|\d]\d{1,}|)\s*(.*?)\s*-->(.*?)<!--\s*ENDEVENT\s*-->/su',$extender_core,$matches,PREG_OFFSET_CAPTURE)
				)
				{
					for ( $i=0; $i<count($matches[0]); $i++ )
					{
						$priority = empty($matches[1][$i][0]) ? 0 : ((int) $matches[1][$i][0]);
						$names = explode(',',str_replace(array(' ',';','|'),array('',',',','),$matches[2][$i][0]));
						foreach( $names as $name )
						{
							if ( empty($name) ) continue;
							if ( !preg_match('/^[a-z0-9_\.]+$/',$name) ) continue;
							self::$core_events[$eindex] = array(
								'name'					=> $name,
								'priority'				=> $priority,
								'exec_source'			=> $matches[3][$i][0],
								'exec_source_addlines'	=> substr_count($extender_core,"\n",0,$matches[3][$i][1]),
								'eval'					=> htmlspecialchars_decode($matches[3][$i][0])
							);
							$ret[$name][] = array('event_extender_core_'.$eindex,$priority);
							$eindex++;
						}
					}
				}
			}
		}

		$ret['core.page_footer_after'][] = array('event_extender_template',-PHP_INT_MAX);
		$ret['core.adm_page_footer'][] = array('event_extender_template',-PHP_INT_MAX);
		$ret['core.common'][] = array('event_extender_set_global_fatal_error_handler',PHP_INT_MAX);

		return $ret;
	}

//-----------------------------------------------------------------------------
//
// from phpBB 3.0.13 functions_template.php
//
	// template
	private $tpl;

	// Various storage arrays
	private $block_names = array();
	private $block_else_level = array();

	/**
	* Remove any PHP tags that do not belong, these regular expressions are derived from
	* the ones that exist in zend_language_scanner.l
	* @access private
	*/
	function remove_php_tags(&$code)
	{
		// This matches the information gathered from the internal PHP lexer
		$match = array(
			'#<([\?%])=?.*?\1>#s',
			'#<script\s+language\s*=\s*(["\']?)php\1\s*>.*?</script\s*>#s',
			'#<\?php(?:\r\n?|[ \n\t]).*?\?>#s'
		);

		$code = preg_replace($match, '', $code);
	}

	/**
	* The all seeing all doing compile method. Parts are inspired by or directly from Smarty
	* @access private
	*/
	function compile($code, $no_echo = false)
	{
		// Remove any "loose" php ... we want to give admins the ability
		// to switch on/off PHP for a given template. Allowing unchecked
		// php is a no-no. There is a potential issue here in that non-php
		// content may be removed ... however designers should use entities
		// if they wish to display < and >
		$this->remove_php_tags($code);

		// Pull out all block/statement level elements and separate plain text
		preg_match_all('#<!-- PHP -->(.*?)<!-- ENDPHP -->#s', $code, $matches);
		$php_blocks = $matches[1];
		$code = preg_replace('#<!-- PHP -->.*?<!-- ENDPHP -->#s', '<!-- PHP -->', $code);

		preg_match_all('#<!-- ([^<].*?) (.*?)? ?-->#', $code, $blocks, PREG_SET_ORDER);

		$text_blocks = preg_split('#<!-- [^<].*? (?:.*?)? ?-->#', $code);

		for ($i = 0, $j = sizeof($text_blocks); $i < $j; $i++)
		{
			$this->compile_var_tags($text_blocks[$i]);
		}
		$compile_blocks = array();

		for ($curr_tb = 0, $tb_size = sizeof($blocks); $curr_tb < $tb_size; $curr_tb++)
		{
			$block_val = &$blocks[$curr_tb];

			switch ($block_val[1])
			{
				case 'BEGIN':
					$this->block_else_level[] = false;
					$compile_blocks[] = '<?php ' . $this->compile_tag_block($block_val[2]) . ' ?>';
				break;

				case 'BEGINELSE':
					$this->block_else_level[sizeof($this->block_else_level) - 1] = true;
					$compile_blocks[] = '<?php }} else { ?>';
				break;

				case 'END':
					array_pop($this->block_names);
					$compile_blocks[] = '<?php ' . ((array_pop($this->block_else_level)) ? '}' : '}}') . ' ?>';
				break;

				case 'IF':
					$compile_blocks[] = '<?php ' . $this->compile_tag_if($block_val[2], false) . ' ?>';
				break;

				case 'ELSE':
					$compile_blocks[] = '<?php } else { ?>';
				break;

				case 'ELSEIF':
					$compile_blocks[] = '<?php ' . $this->compile_tag_if($block_val[2], true) . ' ?>';
				break;

				case 'ENDIF':
					$compile_blocks[] = '<?php } ?>';
				break;

				case 'DEFINE':
					$compile_blocks[] = '<?php ' . $this->compile_tag_define($block_val[2], true) . ' ?>';
				break;

				case 'UNDEFINE':
					$compile_blocks[] = '<?php ' . $this->compile_tag_define($block_val[2], false) . ' ?>';
				break;

				case 'PHP':
					$compile_blocks[] = ($this->allow_php) ? '<?php ' . array_shift($php_blocks) . ' ?>' : '';
				break;

				default:
					$this->compile_var_tags($block_val[0]);
					$trim_check = trim($block_val[0]);
					$compile_blocks[] = (!$no_echo) ? ((!empty($trim_check)) ? $block_val[0] : '') : ((!empty($trim_check)) ? $block_val[0] : '');
				break;
			}
		}

		$tpl_php = '';
		for ($i = 0, $size = sizeof($text_blocks); $i < $size; $i++)
		{
			$trim_check_text = trim($text_blocks[$i]);
			$tpl_php .= (!$no_echo) ? (($trim_check_text != '') ? $text_blocks[$i] : '') . ((isset($compile_blocks[$i])) ? $compile_blocks[$i] : '') : (($trim_check_text != '') ? $text_blocks[$i] : '') . ((isset($compile_blocks[$i])) ? $compile_blocks[$i] : '');
		}

		// Remove unused opening/closing tags
		$tpl_php = str_replace(' ?><?php ', ' ', $tpl_php);

		// Now add a newline after each php closing tag which already has a newline
		// PHP itself strips a newline if a closing tag is used (this is documented behaviour) and it is mostly not intended by style authors to remove newlines
		$tpl_php = preg_replace('#\?\>([\r\n])#', '?>\1\1', $tpl_php);

		return $tpl_php;
	}

	/**
	* Compile variables
	* @access private
	*/
	function compile_var_tags(&$text_blocks)
	{
		// change template varrefs into PHP varrefs
		$varrefs = array();

		// This one will handle varrefs WITH namespaces
		preg_match_all('#\{((?:[a-z0-9\-_]+\.)+)(\$)?([A-Z0-9\-_]+)\}#', $text_blocks, $varrefs, PREG_SET_ORDER);

		foreach ($varrefs as $var_val)
		{
			$namespace = $var_val[1];
			$varname = $var_val[3];
			$new = $this->generate_block_varref($namespace, $varname, true, $var_val[2]);

			$text_blocks = str_replace($var_val[0], $new, $text_blocks);
		}

		// This will handle the remaining root-level varrefs
		// transform vars prefixed by L_ into their language variable pendant if nothing is set within the tpldata array
		if (strpos($text_blocks, '{L_') !== false)
		{
			$text_blocks = preg_replace('#\{L_([A-Z0-9\-_]+)\}#', "<?php echo ((isset(\$this->rootref['L_\\1'])) ? \$this->rootref['L_\\1'] : ((isset(\$this->userlang['\\1'])) ? \$this->userlang['\\1'] : '{ \\1 }')); ?>", $text_blocks);
		}

		// Handle addslashed language variables prefixed with LA_
		// If a template variable already exist, it will be used in favor of it...
		if (strpos($text_blocks, '{LA_') !== false)
		{
			$text_blocks = preg_replace('#\{LA_([A-Z0-9\-_]+)\}#', "<?php echo ((isset(\$this->rootref['LA_\\1'])) ? \$this->rootref['LA_\\1'] : ((isset(\$this->rootref['L_\\1'])) ? addslashes(\$this->rootref['L_\\1']) : ((isset(\$this->userlang['\\1'])) ? addslashes(\$this->userlang['\\1']) : '{ \\1 }'))); ?>", $text_blocks);
		}

		// Handle remaining varrefs
		$text_blocks = preg_replace('#\{([A-Z0-9\-_]+)\}#', "<?php echo (isset(\$this->rootref['\\1'])) ? \$this->rootref['\\1'] : ''; ?>", $text_blocks);
		$text_blocks = preg_replace('#\{\$([A-Z0-9\-_]+)\}#', "<?php echo (isset(\$this->tpldata['DEFINE']['.']['\\1'])) ? \$this->tpldata['DEFINE']['.']['\\1'] : ''; ?>", $text_blocks);

		return;
	}

	/**
	* Compile blocks
	* @access private
	*/
	function compile_tag_block($tag_args)
	{
		$no_nesting = false;

		// Is the designer wanting to call another loop in a loop?
		if (strpos($tag_args, '!') === 0)
		{
			// Count the number of ! occurrences (not allowed in vars)
			$no_nesting = substr_count($tag_args, '!');
			$tag_args = substr($tag_args, $no_nesting);
		}

		// Allow for control of looping (indexes start from zero):
		// foo(2)    : Will start the loop on the 3rd entry
		// foo(-2)   : Will start the loop two entries from the end
		// foo(3,4)  : Will start the loop on the fourth entry and end it on the fifth
		// foo(3,-4) : Will start the loop on the fourth entry and end it four from last
		if (preg_match('#^([^()]*)\(([\-\d]+)(?:,([\-\d]+))?\)$#', $tag_args, $match))
		{
			$tag_args = $match[1];

			if ($match[2] < 0)
			{
				$loop_start = '($_' . $tag_args . '_count ' . $match[2] . ' < 0 ? 0 : $_' . $tag_args . '_count ' . $match[2] . ')';
			}
			else
			{
				$loop_start = '($_' . $tag_args . '_count < ' . $match[2] . ' ? $_' . $tag_args . '_count : ' . $match[2] . ')';
			}

			if (strlen($match[3]) < 1 || $match[3] == -1)
			{
				$loop_end = '$_' . $tag_args . '_count';
			}
			else if ($match[3] >= 0)
			{
				$loop_end = '(' . ($match[3] + 1) . ' > $_' . $tag_args . '_count ? $_' . $tag_args . '_count : ' . ($match[3] + 1) . ')';
			}
			else //if ($match[3] < -1)
			{
				$loop_end = '$_' . $tag_args . '_count' . ($match[3] + 1);
			}
		}
		else
		{
			$loop_start = 0;
			$loop_end = '$_' . $tag_args . '_count';
		}

		$tag_template_php = '';
		array_push($this->block_names, $tag_args);

		if ($no_nesting !== false)
		{
			// We need to implode $no_nesting times from the end...
			$block = array_slice($this->block_names, -$no_nesting);
		}
		else
		{
			$block = $this->block_names;
		}

		if (sizeof($block) < 2)
		{
			// Block is not nested.
			$tag_template_php = '$_' . $tag_args . "_count = (isset(\$this->tpldata['$tag_args'])) ? sizeof(\$this->tpldata['$tag_args']) : 0;";
			$varref = "\$this->tpldata['$tag_args']";
		}
		else
		{
			// This block is nested.
			// Generate a namespace string for this block.
			$namespace = implode('.', $block);

			// Get a reference to the data array for this block that depends on the
			// current indices of all parent blocks.
			$varref = $this->generate_block_data_ref($namespace, false);

			// Create the for loop code to iterate over this block.
			$tag_template_php = '$_' . $tag_args . '_count = (isset(' . $varref . ')) ? sizeof(' . $varref . ') : 0;';
		}

		$tag_template_php .= 'if ($_' . $tag_args . '_count) {';

		/**
		* The following uses foreach for iteration instead of a for loop, foreach is faster but requires PHP to make a copy of the contents of the array which uses more memory
		* <code>
		*	if (!$offset)
		*	{
		*		$tag_template_php .= 'foreach (' . $varref . ' as $_' . $tag_args . '_i => $_' . $tag_args . '_val){';
		*	}
		* </code>
		*/

		$tag_template_php .= 'for ($_' . $tag_args . '_i = ' . $loop_start . '; $_' . $tag_args . '_i < ' . $loop_end . '; ++$_' . $tag_args . '_i){';
		$tag_template_php .= '$_'. $tag_args . '_val = &' . $varref . '[$_'. $tag_args. '_i];';

		return $tag_template_php;
	}

	/**
	* Compile IF tags - much of this is from Smarty with
	* some adaptions for our block level methods
	* @access private
	*/
	function compile_tag_if($tag_args, $elseif)
	{
		// Tokenize args for 'if' tag.
		preg_match_all('/(?:
			"[^"\\\\]*(?:\\\\.[^"\\\\]*)*"         |
			\'[^\'\\\\]*(?:\\\\.[^\'\\\\]*)*\'     |
			[(),]                                  |
			[^\s(),]+)/x', $tag_args, $match);

		$tokens = $match[0];
		$is_arg_stack = array();

		for ($i = 0, $size = sizeof($tokens); $i < $size; $i++)
		{
			$token = &$tokens[$i];

			switch ($token)
			{
				case '!==':
				case '===':
				case '<<':
				case '>>':
				case '|':
				case '^':
				case '&':
				case '~':
				case ')':
				case ',':
				case '+':
				case '-':
				case '*':
				case '/':
				case '@':
				break;

				case '==':
				case 'eq':
					$token = '==';
				break;

				case '!=':
				case '<>':
				case 'ne':
				case 'neq':
					$token = '!=';
				break;

				case '<':
				case 'lt':
					$token = '<';
				break;

				case '<=':
				case 'le':
				case 'lte':
					$token = '<=';
				break;

				case '>':
				case 'gt':
					$token = '>';
				break;

				case '>=':
				case 'ge':
				case 'gte':
					$token = '>=';
				break;

				case '&&':
				case 'and':
					$token = '&&';
				break;

				case '||':
				case 'or':
					$token = '||';
				break;

				case '!':
				case 'not':
					$token = '!';
				break;

				case '%':
				case 'mod':
					$token = '%';
				break;

				case '(':
					array_push($is_arg_stack, $i);
				break;

				case 'is':
					$is_arg_start = ($tokens[$i-1] == ')') ? array_pop($is_arg_stack) : $i-1;
					$is_arg	= implode('	', array_slice($tokens,	$is_arg_start, $i -	$is_arg_start));

					$new_tokens	= $this->_parse_is_expr($is_arg, array_slice($tokens, $i+1));

					array_splice($tokens, $is_arg_start, sizeof($tokens), $new_tokens);

					$i = $is_arg_start;

				// no break

				default:
					if (preg_match('#^((?:[a-z0-9\-_]+\.)+)?(\$)?(?=[A-Z])([A-Z0-9\-_]+)#s', $token, $varrefs))
					{
						$token = (!empty($varrefs[1])) ? $this->generate_block_data_ref(substr($varrefs[1], 0, -1), true, $varrefs[2]) . '[\'' . $varrefs[3] . '\']' : (($varrefs[2]) ? '$this->tpldata[\'DEFINE\'][\'.\'][\'' . $varrefs[3] . '\']' : '$this->rootref[\'' . $varrefs[3] . '\']');
					}
					else if (preg_match('#^\.((?:[a-z0-9\-_]+\.?)+)$#s', $token, $varrefs))
					{
						// Allow checking if loops are set with .loopname
						// It is also possible to check the loop count by doing <!-- IF .loopname > 1 --> for example
						$blocks = explode('.', $varrefs[1]);

						// If the block is nested, we have a reference that we can grab.
						// If the block is not nested, we just go and grab the block from _tpldata
						if (sizeof($blocks) > 1)
						{
							$block = array_pop($blocks);
							$namespace = implode('.', $blocks);
							$varref = $this->generate_block_data_ref($namespace, true);

							// Add the block reference for the last child.
							$varref .= "['" . $block . "']";
						}
						else
						{
							$varref = '$this->tpldata';

							// Add the block reference for the last child.
							$varref .= "['" . $blocks[0] . "']";
						}
						$token = "sizeof($varref)";
					}
					else if (!empty($token))
					{
						$token = '(' . $token . ')';
					}

				break;
			}
		}

		// If there are no valid tokens left or only control/compare characters left, we do skip this statement
		if (!sizeof($tokens) || str_replace(array(' ', '=', '!', '<', '>', '&', '|', '%', '(', ')'), '', implode('', $tokens)) == '')
		{
			$tokens = array('false');
		}
		return (($elseif) ? '} else if (' : 'if (') . (implode(' ', $tokens) . ') { ');
	}

	/**
	* Compile DEFINE tags
	* @access private
	*/
	function compile_tag_define($tag_args, $op)
	{
		preg_match('#^((?:[a-z0-9\-_]+\.)+)?\$(?=[A-Z])([A-Z0-9_\-]*)(?: = (\'?)([^\']*)(\'?))?$#', $tag_args, $match);

		if (empty($match[2]) || (!isset($match[4]) && $op))
		{
			return '';
		}

		if (!$op)
		{
			return 'unset(' . (($match[1]) ? $this->generate_block_data_ref(substr($match[1], 0, -1), true, true) . '[\'' . $match[2] . '\']' : '$this->tpldata[\'DEFINE\'][\'.\'][\'' . $match[2] . '\']') . ');';
		}

		// Are we a string?
		if ($match[3] && $match[5])
		{
			$match[4] = str_replace(array('\\\'', '\\\\', '\''), array('\'', '\\', '\\\''), $match[4]);

			// Compile reference, we allow template variables in defines...
			$match[4] = $this->compile($match[4]);

			// Now replace the php code
			$match[4] = "'" . str_replace(array('<?php echo ', '; ?>'), array("' . ", " . '"), $match[4]) . "'";
		}
		else
		{
			preg_match('#true|false|\.#i', $match[4], $type);

			switch (strtolower($type[0]))
			{
				case 'true':
				case 'false':
					$match[4] = strtoupper($match[4]);
				break;

				case '.':
					$match[4] = doubleval($match[4]);
				break;

				default:
					$match[4] = intval($match[4]);
				break;
			}
		}

		return (($match[1]) ? $this->generate_block_data_ref(substr($match[1], 0, -1), true, true) . '[\'' . $match[2] . '\']' : '$this->tpldata[\'DEFINE\'][\'.\'][\'' . $match[2] . '\']') . ' = ' . $match[4] . ';';
	}

	/**
	* parse expression
	* This is from Smarty
	* @access private
	*/
	function _parse_is_expr($is_arg, $tokens)
	{
		$expr_end = 0;
		$negate_expr = false;

		if (($first_token = array_shift($tokens)) == 'not')
		{
			$negate_expr = true;
			$expr_type = array_shift($tokens);
		}
		else
		{
			$expr_type = $first_token;
		}

		switch ($expr_type)
		{
			case 'even':
				if (@$tokens[$expr_end] == 'by')
				{
					$expr_end++;
					$expr_arg = $tokens[$expr_end++];
					$expr = "!(($is_arg / $expr_arg) % $expr_arg)";
				}
				else
				{
					$expr = "!($is_arg & 1)";
				}
			break;

			case 'odd':
				if (@$tokens[$expr_end] == 'by')
				{
					$expr_end++;
					$expr_arg = $tokens[$expr_end++];
					$expr = "(($is_arg / $expr_arg) % $expr_arg)";
				}
				else
				{
					$expr = "($is_arg & 1)";
				}
			break;

			case 'div':
				if (@$tokens[$expr_end] == 'by')
				{
					$expr_end++;
					$expr_arg = $tokens[$expr_end++];
					$expr = "!($is_arg % $expr_arg)";
				}
			break;
		}

		if ($negate_expr)
		{
			$expr = "!($expr)";
		}

		array_splice($tokens, 0, $expr_end, $expr);

		return $tokens;
	}

	/**
	* Generates a reference to the given variable inside the given (possibly nested)
	* block namespace. This is a string of the form:
	* ' . $this->tpldata['parent'][$_parent_i]['$child1'][$_child1_i]['$child2'][$_child2_i]...['varname'] . '
	* It's ready to be inserted into an "echo" line in one of the templates.
	* NOTE: expects a trailing "." on the namespace.
	* @access private
	*/
	function generate_block_varref($namespace, $varname, $echo = true, $defop = false)
	{
		// Strip the trailing period.
		$namespace = substr($namespace, 0, -1);

		// Get a reference to the data block for this namespace.
		$varref = $this->generate_block_data_ref($namespace, true, $defop);
		// Prepend the necessary code to stick this in an echo line.

		// Append the variable reference.
		$varref .= "['$varname']";
		$varref = ($echo) ? "<?php echo $varref; ?>" : ((isset($varref)) ? $varref : '');

		return $varref;
	}

	/**
	* Generates a reference to the array of data values for the given
	* (possibly nested) block namespace. This is a string of the form:
	* $this->tpldata['parent'][$_parent_i]['$child1'][$_child1_i]['$child2'][$_child2_i]...['$childN']
	*
	* If $include_last_iterator is true, then [$_childN_i] will be appended to the form shown above.
	* NOTE: does not expect a trailing "." on the blockname.
	* @access private
	*/
	function generate_block_data_ref($blockname, $include_last_iterator, $defop = false)
	{
		// Get an array of the blocks involved.
		$blocks = explode('.', $blockname);
		$blockcount = sizeof($blocks) - 1;

		// DEFINE is not an element of any referenced variable, we must use _tpldata to access it
		if ($defop)
		{
			$varref = '$this->tpldata[\'DEFINE\']';
			// Build up the string with everything but the last child.
			for ($i = 0; $i < $blockcount; $i++)
			{
				$varref .= "['" . $blocks[$i] . "'][\$_" . $blocks[$i] . '_i]';
			}
			// Add the block reference for the last child.
			$varref .= "['" . $blocks[$blockcount] . "']";
			// Add the iterator for the last child if requried.
			if ($include_last_iterator)
			{
				$varref .= '[$_' . $blocks[$blockcount] . '_i]';
			}
			return $varref;
		}
		else if ($include_last_iterator)
		{
			return '$_'. $blocks[$blockcount] . '_val';
		}
		else
		{
			return '$_'. $blocks[$blockcount - 1] . '_val[\''. $blocks[$blockcount]. '\']';
		}
	}

//-----------------------------------------------------------------------------
//
// Set global fatal error handler
//
	public function event_extender_set_global_fatal_error_handler($event)
	{
		require($this->phpbb_root_path.'ext/c61/extender/includes/extender_fatal_error_handler.'.$this->php_ext);
	}

//-----------------------------------------------------------------------------

	// Shutdown
	private $shutdown_type = array();
	private $exec_type = '';
	private $lock_message_sent = false;
	private $locked = false;
	private $lockfile = '';
	private $error_last = null;
	private $handler_error = null;
	private $e_errors = array(
			E_ERROR				=> 'E_ERROR',
			E_WARNING			=> 'E_WARNING',
			E_PARSE				=> 'E_PARSE',
			E_NOTICE			=> 'E_NOTICE',
			E_CORE_ERROR		=> 'E_CORE_ERROR',
			E_CORE_WARNING		=> 'E_CORE_WARNING',
			E_COMPILE_ERROR		=> 'E_COMPILE_ERROR',
			E_COMPILE_WARNING	=> 'E_COMPILE_WARNING',
			E_USER_ERROR		=> 'E_USER_ERROR',
			E_USER_WARNING		=> 'E_USER_WARNING',
			E_USER_NOTICE		=> 'E_USER_NOTICE',
			E_STRICT			=> 'E_STRICT',
			E_RECOVERABLE_ERROR	=> 'E_RECOVERABLE_ERROR',
			E_DEPRECATED		=> 'E_DEPRECATED',
			E_USER_DEPRECATED	=> 'E_USER_DEPRECATED',
	);

	// debug mode
	private $debug = true;
	private $debug_warnings = true;

	// listing (if debug and compiling errors or fatal error)
	private $listing = '';
	// listing pre/succ
	private $listing_pre = "<div width=\"100%\" style=\"background-color:white; padding:10px; text-align:left;\"><font color=\"black\" face=\"monospace\" size=\"3\"><b style=\"color:darkred;\">Extender information:</b><br />\n";
	private $listing_succ = "<br />\n</font></div>";
	private $listing_spre = "<div width=\"100%\" style=\"background-color:white; padding:10px; text-align:left;\"><font color=\"darkred\" face=\"monospace\" size=\"3\"><b>";
	private $listing_ssucc = "</b><br />\n</font></div>";

	// tpl done ?
	private $event_template_done = array();

	// hashes
	private $hash = array();

	// lock filename (at sys temp dir)
	private function lockfilename($type)
	{
		$this->lockfile = sys_get_temp_dir().DIRECTORY_SEPARATOR.'Extender_'.dechex(crc32($this->config['cookie_domain'].$this->config['cookie_name'])).'_'.$type.'.lock';
	}

	// shutdown function for fatal errors
	public function extender_shutdown_function($type)
	{
		$error = error_get_last();
		if ( $error )
		{
			global $extender_fatal_error_handler_called, $extender_fatal_error_handler_active;

			if ( in_array($type,$this->shutdown_type)
			  && ($type == $this->exec_type)
			)
			{
				switch ( $error['type'] )
				{
					case E_ERROR:				// 1
					case E_PARSE:				// 4
					case E_CORE_ERROR:			// 16
					case E_COMPILE_ERROR:		// 64
					case E_USER_ERROR:			// 256
					case E_RECOVERABLE_ERROR:	// 4096
					{
						if ( !headers_sent() ) @header( 'Content-type: text/html; charset=utf-8' );
						echo '<!DOCTYPE html><html dir="ltr" lang="ru"><head><meta charset="utf-8" /></head><body>';

						$fatalmessage = 'Extender '.$type.' FATAL error ['.$this->e_errors[$error['type']].' = '.$error['type'].']: '.$error['message'] .' at file '.phpbb_filter_root_path($error['file']).'('.$error['line'].')';
						echo $this->listing_pre.$fatalmessage.$this->listing_succ;

						$error['fatal'] = true;
						$this->listing($error,'error');
						$message = 'Extender locked, for unlock remove file sys_get_temp_dir/'.basename($this->lockfile);

						file_put_contents($this->lockfile,$this->listing_pre.$this->listing.$this->listing_succ.$this->listing_spre.$message.$this->listing_ssucc);

						echo $this->listing_spre.'Refresh this page now, please... Use F5 in desktop browsers.'.$this->listing_ssucc;
						echo '<hr />';

						echo $this->listing_pre.$this->listing.$this->listing_succ;
						echo $this->listing_spre.$message.$this->listing_ssucc;

						echo '</body></html>';

						global $phpbb_log;

						$user_ip = (empty($this->user->ip)) ? '' : $this->user->ip;
						$log_operation = 'LOG_EXTENDER_FATAL_ERROR';
						$data = '<br />'.$this->exec_event_type.' '.$this->exec_event_name.'('.($this->exec_source_addlines ? (($this->exec_source_addlines+$error['line']).'/') : '').$error['line'].')'.' » ['.$this->e_errors[$error['type']].' = '.$error['type'].'] '.htmlspecialchars($error['message']);

						$phpbb_log->add('critical', isset($this->user->data['user_id']) ? isset($this->user->data['user_id']) : ANONYMOUS, 	$user_ip, $log_operation, time(), array($data));

						break;
					}
					default:
					{
					    break;
					}
				}
			}
			else if ( ($this->error_last !== null) && ($this->error_last != $error) && !$extender_fatal_error_handler_called )
			{
				// unknown error ? NO white pages !!! let's print message if fatal error
				switch ( $error['type'] )
				{
					case E_ERROR:				// 1
					case E_PARSE:				// 4
					case E_CORE_ERROR:			// 16
					case E_COMPILE_ERROR:		// 64
					case E_USER_ERROR:			// 256
					case E_RECOVERABLE_ERROR:	// 4096
					{
						if ( !headers_sent() ) @header( 'Content-type: text/html; charset=utf-8' );
						echo '<!DOCTYPE html><html dir="ltr" lang="ru"><head><meta charset="utf-8" /></head><body>';
			
						$fatalmessage = 'Extender detect FATAL error ['.$this->e_errors[$error['type']].' = '.$error['type'].']: '.$error['message'] .' at file '.phpbb_filter_root_path($error['file']).'('.$error['line'].')';
						echo $this->listing_spre.$fatalmessage.$this->listing_ssucc;

						echo '</body></html>';

						break;
					}
					default:
					{
					    break;
					}
				}
			}
			$extender_fatal_error_handler_called = true;
		}
	}

	// set shutdown
	private function set_shutdown_function($type)
	{
		global $extender_fatal_error_handler_active;

		if ( !in_array($type,$this->shutdown_type) )
		{
			$this->shutdown_type[] = $type;
			register_shutdown_function(array($this,'extender_shutdown_function'),$type);
		}
		$extender_fatal_error_handler_active = true;
	}

	// remove shutdown (fictive)
	private function remove_shutdown_function($type)
	{
		global $extender_fatal_error_handler_active;

		$extender_fatal_error_handler_active = false;
	}

	// error handler
	private function error_handler($errno, $errstr, $errfile, $errline)
	{
		if ( $this->handler_error === null )
		{
			$this->handler_error = array(
				'type'		=> $errno,
				'message'	=> $errstr,
				'file'		=> $errfile,
				'line'		=> $errline,
			);
		}
		return false;
	}

	// lock message
	private function lock_message($type)
	{
		if( $this->locked )
		{
			return true;
		}
		else if ( $this->lock_message_sent )
		{
			return false;
		}
		else
		{
			$this->lockfilename($type);
			$this->lock_message_sent = true;
			if ( file_exists($this->lockfile) )
			{
				if ( $this->debug )
				{
					$this->template->assign_vars(array(
						'S_EXTENDER_LISTING'	=> file_get_contents($this->lockfile),
					));
				}
				$this->locked = true;
				return true;
			}
		}
		return false;
	}

	// Get context
	private function get_context()
	{
		// rootref & template data context
		global $phpbb_container;

		$context = $phpbb_container->get('template_context');
		$this->rootref = &$context->get_root_ref();
		$this->tpldata = &$context->get_data_ref();

		// user lang
		$this->userlang = &$this->user->lang;

		// Modes
		if ( isset($this->config['extender_debug']) ) $this->debug = $this->config['extender_debug'];
		if ( isset($this->config['extender_debug_warnings']) ) $this->debug_warnings = $this->config['extender_debug_warnings'];
	}

	// Listing
	private function listing($error,$ew)
	{
		$tpl_list_ary = explode("\n",htmlspecialchars($this->exec_source));
		$this->listing .=
			 '<br /><b style="color:darkred;">'.$this->exec_event_type.' '.$this->exec_event_name.'('.($this->exec_source_addlines ? (($this->exec_source_addlines+$error['line']).'/') : '').$error['line'].')</b>'
			.'<b>: '.(isset($error['fatal']) ? 'FATAL ' : '').$ew.' ['.$this->e_errors[$error['type']].' = '.$error['type'].'] '.htmlspecialchars($error['message']).' at file: '.phpbb_filter_root_path($error['file']).'</b>'
			.'<pre>';
		$tpl_list_ary[0] = '<i>/*&lt;!-- EVENT '.$this->exec_event_name.' --&gt;*/</i>' . $tpl_list_ary[0];
		$tpl_list_ary[count($tpl_list_ary)-1] = trim($tpl_list_ary[count($tpl_list_ary)-1]) . '<i>/*&lt;!-- ENDEVENT --&gt;</i>*/';
		$l = 1;
		foreach ( $tpl_list_ary as $tpl_list )
		{
			if ( $this->exec_source_addlines )
			{
				$this->listing .= sprintf("%04d/%04d",$l+$this->exec_source_addlines,$l).' : '.$tpl_list."\n";
				if ( $error['line'] == $l ) $this->listing .= '<b>**** **** ('.($this->exec_source_addlines ? (($this->exec_source_addlines+$error['line']).'/') : '').$error['line'].') '.(isset($error['fatal']) ? 'FATAL ' : '').$ew.' ['.$this->e_errors[$error['type']].' = '.$error['type'].'] '.htmlspecialchars($error['message'])."</b>\n";
			}
			else
			{
				$this->listing .= sprintf("%04d",$l).' : '.$tpl_list."\n";
				if ( $error['line'] == $l ) $this->listing .= '<b>****  ('.$error['line'].') '.(isset($error['fatal']) ? 'FATAL ' : '').$ew.' ['.$this->e_errors[$error['type']].' = '.$error['type'].'] '.htmlspecialchars($error['message'])."</b>\n";
			}
			$l++;
		}
		$this->listing .= '</pre>';
	}

	// Execute core/tpl (eval)
	private function exec_eval()
	{
		// restore local vars
		if ( $this->saved_vars !== null ) extract($this->saved_vars);

		// eval and get errors
		if ( (@eval($this->eval) === false)
		  && $this->debug
		  && ($error = error_get_last())
		)
		{
			return $error;
		}

		// save local vars
		$this->saved_vars = get_defined_vars();
		// if register globals ON, but it's generate warning
//		$this->saved_vars = @array_diff(get_defined_vars(), array(array()));

		return false;
	}
	
	/**
	* Events
	*/
	public function __call($func,$args)
	{
		if ( strpos($func,'event_extender_core_') === 0 )
		{
			$this->event_extender_core($args[0],$func);
		}
	}
	public function event_extender_core($event,$efunction = '')
	{
		// Core events index
		if ( empty($efunction) ) return;
		$eindex = (int) str_replace('event_extender_core_','',$efunction);
		if ( !isset(self::$core_events[$eindex]) ) return;

		// Get context
		$this->get_context();
		$this->event = &$event;
		
		// Shutdown
		$type = 'Core';
		$this->set_shutdown_function($type);
		if ( $this->lock_message($type) )
		{
			$this->remove_shutdown_function($type);
			return;
		}
		
		// Event Name
		$eventName = $event->getName();
		
		// Event
		if ( isset($this->config['extender_core_on']) && $this->config['extender_core_on'] )
		{
			$this->exec_type = $type;
			$this->exec_event_type = $type.' Event';
			$this->exec_event_name = $eventName;
			$this->exec_source = &self::$core_events[$eindex]['exec_source'];
			$this->exec_source_addlines = self::$core_events[$eindex]['exec_source_addlines'];
			$this->eval = &self::$core_events[$eindex]['eval'];
			
			$errorlevel = error_reporting(-1);
			$errorhandler = set_error_handler(array($this,'error_handler'),E_NOTICE | E_WARNING | E_DEPRECATED | E_STRICT | E_USER_DEPRECATED | E_USER_WARNING | E_USER_NOTICE);
			$display_errors = function_exists('ini_get') ? @ini_get('display_errors') : false;
			if ( function_exists('ini_set') ) @ini_set('display_errors', '1');
			if ( ($error = $this->exec_eval()) !== false )
			{
				$this->listing($error,'error');
				$this->error_last = $error;
			}
			if ( ($this->handler_error != null)
			  && $this->debug_warnings
			)
			{
				$this->listing($this->handler_error,'warning');
				$this->handler_error = null;
			}
			if ( function_exists('ini_set') && ($display_errors !== false) ) @ini_set('display_errors', $display_errors);
			restore_error_handler();
			error_reporting($errorlevel);
			
			$this->exec_type = '';
		}

		// Shutdown
		$this->remove_shutdown_function($type);
	}

	public function event_extender_template($event,$eventName = '')
	{
		// Get context
		$this->get_context();
		$this->event = &$event;

		// Cookie
		$this->template->assign_var('S_EXTENDER_COOKIENAME',$this->config['cookie_name'] . '_extdr');

		// Shutdown
		$type = 'Template';
		$this->set_shutdown_function($type);
		if ( $this->lock_message($type) )
		{
			$this->remove_shutdown_function($type);
			return;
		}
		
		// ACP ?
		$acp = defined('ADMIN_START') ? true : false;

		// Template
		if ( isset($this->config['extender_template_on']) && $this->config['extender_template_on'] )
		{
			$extender_template_data = $this->config_text->get_array(array('extender_template'));
			$extender_template = htmlspecialchars_decode($extender_template_data['extender_template']);
			if ( !empty($extender_template)
			  && preg_match_all('/<!--\s*EVENT\s*(PRE\s*|)(.*?)\s*-->(.*?)<!--\s*ENDEVENT\s*-->/su',$extender_template,$matches,PREG_OFFSET_CAPTURE)
			)
			{
				for ( $i=0; $i<count($matches[0]); $i++ )
				{
					$name = &$matches[2][$i][0];
					if ( !preg_match('/^[a-z0-9_]+$/',$name) ) continue;
					$name_acp = (strpos($name,'acp_') === 0) ? true : false;
					if ( $name_acp )
					{
						if ( !$acp ) continue;
					}
					else
					{
						if ( $acp ) continue;
					}
					$varname = 'S_EXTENDER_'.strtoupper($name);
					$pre = (stripos($matches[1][$i][0],'PRE') === 0) ? true : false;
					$source = &$matches[3][$i][0];
						
					$this->hash[$varname] = sprintf('%u',dechex(crc32((isset($this->hash[$varname]) ? $this->hash[$varname] : '').$varname.$source)));
					$this->template->assign_var($varname.'_HASH',$this->hash[$varname]);

					if ( $pre )
					{
						ob_start();

						$this->eval = '?>'.$this->compile(htmlspecialchars_decode($source));
						$this->exec_type = $type;
						$this->exec_event_type = ($name_acp ? 'ACP ' : '').$type.' Event';
						$this->exec_event_name = ($pre ? 'PRE ' : '') . $name;
						$this->exec_source = &$source;
						$this->exec_source_addlines = substr_count($extender_template,"\n",0,$matches[3][$i][1]);

						$errorlevel = error_reporting(-1);
						$errorhandler = set_error_handler(array($this,'error_handler'),E_NOTICE | E_WARNING | E_DEPRECATED | E_STRICT | E_USER_DEPRECATED | E_USER_WARNING | E_USER_NOTICE);
						$display_errors = function_exists('ini_get') ? @ini_get('display_errors') : false;
						if ( function_exists('ini_set') ) @ini_set('display_errors', '1');
						
						if ( ($error = $this->exec_eval()) !== false )
						{
							$this->listing($error,'error');
							$this->error_last = $error;
						}
						if ( ($this->handler_error != null)
						  && $this->debug_warnings
						)
						{
							$this->listing($this->handler_error,'warning');
							$this->handler_error = null;
						}
						if ( function_exists('ini_set') && ($display_errors !== false) ) @ini_set('display_errors', $display_errors);
						restore_error_handler();
						error_reporting($errorlevel);

						$this->exec_type = '';

						$source = ob_get_contents();
						ob_end_clean();
					}
					else
					{
						$source = '';
					}

					$this->template->append_var($varname, $source);
				}
			}
		}
		
		// Shutdown
		$this->remove_shutdown_function($type);

		// Debug & listing
		if ( $this->debug
		  && !empty($this->listing)
		)
		{
			$this->template->assign_vars(array(
				'S_EXTENDER_LISTING'	=> $this->listing_pre.$this->listing.$this->listing_succ,
			));
		}

		// Done.
		if ( !empty($eventName) )
		{
			$this->event_template_done[$eventName] = true;
		}
	}

}
